﻿using FObjBase;
using FLY.Thick.Blowing.Common;
using FLY.Thick.Blowing.IService;
using FLY.Thick.Blowing.Server.Model;
using LiveCharts;
using LiveCharts.Configurations;
using LiveCharts.Wpf;
using Misc;
using NLog.Config;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Data;
using System.Windows.Media;

namespace FLY.Thick.Blowing.UI.Fix.Client.UiModule
{
    public class ScanGraphVm : INotifyPropertyChanged
    {
        const int MARKNO_TEXTUP = 1;
        const int MARKNO_TARGET_DELAY = 2;
        #region 图表控制
        public double YMax { get; set; } = double.NaN;
        public double YMin { get; set; } = double.NaN;
        public double Target { get; set; } = double.NaN;
        public double ToleranceYMax { get; set; } = double.NaN;
        public double ToleranceYMin { get; set; } = double.NaN;
        public double Tolerance2YMax { get; set; } = double.NaN;
        public double Tolerance2YMin { get; set; } = double.NaN;
        public double XMax { get; set; } = double.NaN;
        public double XMin { get; set; } = double.NaN;
        public int XStep { get; set; } = 5;
        public Func<double, string> YFormatter { get; set; }

        public SeriesCollection Series { get; set; }

        #endregion



        #region 界面统计值
        /// <summary>
        /// 平均值
        /// </summary>
        public double Average { get; private set; } = double.NaN;

        public double Sigma2 { get; private set; } = double.NaN;
        /// <summary>
        /// 最大值
        /// </summary>
        public double Max { get; private set; } = double.NaN;
        /// <summary>
        /// 最小值
        /// </summary>
        public double Min { get; private set; } = double.NaN;

        public string Sigma2Text { get; private set; }
        public string MaxText { get; private set; }
        public string MinText { get; private set; }

        /// <summary>
        /// 混合数
        /// </summary>
        public int Mix { get; private set; } = 2;

        /// <summary>
        /// 测量时间
        /// </summary>
        public DateTime Time { get; private set; }

        /// <summary>
        /// 旋转方向 是反向
        /// </summary>
        public bool IsBackw { get; private set; }

        /// <summary>
        /// 旋转1周的时间
        /// </summary>
        public TimeSpan RPeriod { get; private set; }


        /// <summary>
        /// 数据库中的序号
        /// </summary>
        public long Id { get; private set; } = 1231221;
        public int OrgBoltNo { get; private set; } = -100;

        #endregion

        List<Brush> AreaColors;

        ChartValues<double> Values = new ChartValues<double>();

        object Mapper;
        BlowingFixProfileParam profileParam;
        ScanGraphItemParam graphparam;
        IBulkDbService bulkDb;
        Lc_ScanData scanData;
        Db_Profile profile;
        IBlowingService blowingService;
        ParamDictionary paramDictionary;
        public ScanGraphVm()
        {
            #region 与数据无关界面参数


            AreaColors = new List<Brush>();
            AreaColors.AddRange(FLY.ControlLibrary.Themes.Styles.AreaColors);

            YFormatter = (y) =>
            {
                //只显示 Target,ToleranceYMax,ToleranceYMin,Tolerance2YMax,Tolerance2YMin
                if (graphparam.IsPercent)
                {
                    double target = Target;//Average
                    if (target > 0)
                    {
                        if (y == target)
                        {
                            return y.ToString("F1");
                        }
                        else
                        {
                            double percent = 100.0 * (y - target) / target;
                            if (percent > 0)
                            {
                                return $"+{Math.Abs(percent):F1}%";
                            }
                            else
                            {
                                return $"-{Math.Abs(percent):F1}%";
                            }
                        }
                    }
                    else
                    {
                        return y.ToString("F1");
                    }
                }
                else
                {
                    return y.ToString("F1");
                }
            };

            Mapper = Mappers.Xy<double>()
                .X((value, index) =>
                {
                    return index + 1;
                })
                .Y(value => value)
                .Fill((value) =>
                {
                    if (value > Tolerance2YMax)
                    {
                        return AreaColors[0];
                    }
                    else if (value > ToleranceYMax)
                    {
                        return AreaColors[1];
                    }
                    else if (value >= ToleranceYMin)
                    {
                        return AreaColors[2];
                    }
                    else if (value >= Tolerance2YMin)
                    {
                        return AreaColors[1];
                    }
                    else
                    {
                        return AreaColors[0];
                    }
                });
            #endregion

            Series = new SeriesCollection
            {
                new Column2Series
                {
                    Title = "扫描",
                    Values = Values,
                    Stroke = new SolidColorBrush(Colors.Black),
                    StrokeThickness = 1,
                    PointGeometry = null,
                    Configuration = Mapper
                }
            };
            ((Series)Series[0]).SetBinding(Column2Series.YAxisCrossingProperty, new Binding(nameof(Target)) { Source = this });
        }

        public void Init(
            ScanGraphItemParam graphparam,
            IBulkDbService bulkDb,
            IBlowingService blowingService,
            ParamDictionary paramDictionary,
            BlowingFixProfileParam profileParam
            )
        {
            this.graphparam = graphparam;
            this.bulkDb = bulkDb;
            this.blowingService = blowingService;
            this.paramDictionary = paramDictionary;
            this.profileParam = profileParam;
            Init();
        }
        void Init()
        {
            Misc.BindingOperations.SetBinding(graphparam, nameof(graphparam.Mix), this, nameof(Mix));
            paramDictionary.SetBinding(this, nameof(XStep), 5);

            UpdateY();
            UpdateX();

            Values.Clear();

            Id = -1;
            UpdateId();

            bulkDb.PropertyChanged += BulkDb_PropertyChanged;
            graphparam.PropertyChanged += Graphparam_PropertyChanged;
            profileParam.PropertyChanged += ProfileParam_PropertyChanged;
            this.PropertyChanged += ScanGraphVm_PropertyChanged;
        }

        static string[] propertyname_updateMaxMinText = new string[] { 
            nameof(Average),nameof(Max),nameof(Min),nameof(Sigma2)};
        
        private void ScanGraphVm_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {

            if (e.PropertyName == nameof(Target))//nameof(Average)
            {
                if (graphparam.IsPercent)
                {
                    NotifyPropertyChanged(nameof(YFormatter));
                }
            }
            
            if(propertyname_updateMaxMinText.Contains(e.PropertyName))
                UpdateMaxMinText();
            
        }

        private void ProfileParam_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(profileParam.Target))
            {
                UpdateY();
            }
            else if (e.PropertyName == nameof(profileParam.TolerancePercent)) 
            {
                UpdateY();
            }
        }

        private void BulkDb_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(bulkDb.LastScanDataId))
            {
                UpdateId();
            }
            else if (e.PropertyName == nameof(FObjBase.FObjServiceClient.IsConnected))
            {
                if ((bulkDb as FObjBase.FObjServiceClient).IsConnected)
                {
                    UpdateId();
                }
            }
        }

        private void Graphparam_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(graphparam.Number))
            {
                UpdateId();
            }
            else if (e.PropertyName == nameof(graphparam.Mix))
            {
                UpdateId();
            }
            else if (e.PropertyName == nameof(graphparam.YRangePercent))
            {
                UpdateY();
            }
            else if (e.PropertyName == nameof(graphparam.IsPercent))
            {
                NotifyPropertyChanged(nameof(YFormatter));
                UpdateY();
                UpdateMaxMinText();

            }
            else if (e.PropertyName == nameof(graphparam.IsAutoTarget))
            {
                UpdateY();
            }
        }



        void UpdateY()
        {
            if (this.profile != null)
            {
                if (this.bulkDb.LastProfileId != this.profile.ID)
                {
                    //使用 profile 数据显示
                    UpdateY_profile();
                }
                else
                {
                    //使用 profileParam 的参数显示
                    UpdateY_profileParam();
                }
            }
            else {
                //使用 profileParam 的参数显示
                UpdateY_profileParam();
            }




        }
        void UpdateY_profile() 
        {
            double target = this.profile.Target;

            if (graphparam.IsPercent || graphparam.IsAutoTarget)
            {
                if (!double.IsNaN(Average))
                    target = Average;
            }

            double tolerance = this.profile.TolerancePercent * target;

            Target = target;
            ToleranceYMax = target + tolerance;
            ToleranceYMin = target - tolerance;
            Tolerance2YMax = target + tolerance * 2;
            Tolerance2YMin = target - tolerance * 2;

            YMax = target + tolerance * graphparam.YRangePercent;
            YMin = target - tolerance * graphparam.YRangePercent;
        }
        void UpdateY_profileParam() {
            double target = this.profileParam.Target;

            if (graphparam.IsPercent || graphparam.IsAutoTarget)
            {
                if(!double.IsNaN(Average))
                    target = Average;
            }

            double tolerance = this.profileParam.TolerancePercent * target;

            Target = target;
            ToleranceYMax = target + tolerance;
            ToleranceYMin = target - tolerance;
            Tolerance2YMax = target + tolerance * 2;
            Tolerance2YMin = target - tolerance * 2;

            YMax = target + tolerance * graphparam.YRangePercent;
            YMin = target - tolerance * graphparam.YRangePercent;
        }
        
        void UpdateX()
        {
            XMin = 0;
            XMax = this.blowingService.NBolts + 1;
        }

        void UpdateId()
        {
            Update4IdChanged();
        }

        void UpdateMaxMinText()
        {
            PollModule.Current.Poll_JustOnce(() =>
            {
                if (graphparam.IsPercent && (Average > 0))
                {
                    MaxText = $"+{((Max - Average) / Average) * 100.0:F1}%";
                    MinText = $"-{((Average - Min) / Average) * 100.0:F1}%";
                    Sigma2Text = $"{(Sigma2 / Average) * 100.0:F1}%";
                }
                else
                {
                    MaxText = $"{Max:F1}";
                    MinText = $"{Min:F1}";
                    Sigma2Text = $"{Sigma2:F1}";
                }
            }, this, MARKNO_TEXTUP);
        }

        void UpdateAverage() {
            if (this.scanData == null)
            {
                Time = DateTime.MinValue;
                Id = -1;
                OrgBoltNo = -100;
                return;
            }
            else
            {
                Time = this.scanData.Time;
                Id = this.scanData.ID;
                OrgBoltNo = this.scanData.OrgBoltNo;
                IsBackw = this.scanData.IsBackw;
                RPeriod = this.scanData.RPeriod;
            }

            double[] values = this.scanData.Thicks;

            var list = from v in values where !double.IsNaN(v) select v;

            if (list.Count() > 0)
            {
                Average = list.Average();
                Max = list.Max();
                Min = list.Min();
                Sigma2 = list.Sigma() * 2;
            }
            else
            {
                Average = double.NaN;
                Max = double.NaN;
                Min = double.NaN;
                Sigma2 = double.NaN;
            }

        }
        void Update4IdChanged2()
        {
            Values.Clear();
            if (this.scanData == null)
            {
                Time = DateTime.MinValue;
                Id = -1;
                OrgBoltNo = -100;
                return;
            }
            else
            {
                Time = this.scanData.Time;
                Id = this.scanData.ID;
                OrgBoltNo = this.scanData.OrgBoltNo;
            }
            

            
            double[] values = this.scanData.Thicks;

            var list = from v in values where !double.IsNaN(v) select v;

            if (list.Count() > 0)
            {
                Average = list.Average();
                Max = list.Max();
                Min = list.Min();
                Sigma2 = list.Sigma() * 2;
            }
            else
            {
                Average = double.NaN;
                Max = double.NaN;
                Min = double.NaN;
                Sigma2 = double.NaN;
            }

            //必须先更新Y轴, X轴资料，最后再画线！！！！！！！
            Values.AddRange(values);


        }
        void Update4IdChanged()
        {
            if (graphparam.Number > 0
                && bulkDb.LastScanDataId == Id
                && Mix == graphparam.Mix)
            {
                //不需要更新！！！！！！！
                return;
            }
            bulkDb.GetFrame(
                new Pack_GetFrameRequest() { Id = graphparam.Number, Mix = graphparam.Mix },
                (object AsyncContext, object retData) =>
                {
                    Pack_GetFrameReponse reponse = retData as Pack_GetFrameReponse;

                    //if (reponse.scanData != null)
                    {
                        this.scanData = reponse.scanData;
                        this.profile = reponse.profile;
                    }

                    //清除全部数据
                    Values.Clear();
                    //更新平均值,它会引起 YFormatter 刷新
                    UpdateAverage();
                    //更新Target
                    UpdateY();

                    UpdateX();

                    //必须先更新Y轴, X轴资料，最后再画线！！！！！！！
                    //不然 YAxisCrossing 会有机会无效

                    if(scanData != null && scanData.Thicks!=null)
                        Values.AddRange(scanData.Thicks);


                }, this);
        }

        void NotifyPropertyChanged(string propertyName) {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
        public event PropertyChangedEventHandler PropertyChanged;
    }
}
