﻿using FObjBase;
using FLY.Thick.Blowing.Common;
using FLY.Thick.Blowing.IService;
using FLY.Thick.Blowing.Server.Model;
using LiveCharts;
using LiveCharts.Configurations;
using LiveCharts.Wpf;
using Misc;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Data;
using System.Windows.Media;
using Unity.Injection;
using OfficeOpenXml.FormulaParsing.Excel.Functions.Information;

namespace FLY.Thick.Blowing.UI.Fix.Client.UiModule
{
    public class ScanGraph2Vm : INotifyPropertyChanged
    {
        const int MARKNO_TEXT = 1;
        const int MARKNO_Y = 2;
        const int MARKNO_TARGET = 3;
        #region 图表控制
        public double YMax { get; set; } = double.NaN;
        public double YMin { get; set; } = double.NaN;
        public double Target { get; private set; } = double.NaN;

        /// <summary>
        /// 公差%
        /// </summary>
        public double TolerancePercent { get; private set; } = 0.05;
        /// <summary>
        /// 原始平均值
        /// </summary>
        public double OrgTarget { get; private set; } = 100;


        public double ToleranceYMax { get; set; } = double.NaN;
        public double ToleranceYMin { get; set; } = double.NaN;
        public double Tolerance2YMax { get; set; } = double.NaN;
        public double Tolerance2YMin { get; set; } = double.NaN;
        public double XMax { get; set; } = double.NaN;
        public double XMin { get; set; } = double.NaN;
        public int XStep { get; set; } = 5;
        public Func<double, string> YFormatter { get; set; }

        public SeriesCollection Series { get; set; }

        #endregion



        #region 界面统计值
        /// <summary>
        /// 平均值
        /// </summary>
        public double Average { get; private set; } = double.NaN;
        public double Average2 { get; private set; } = double.NaN;

        public double Sigma2 { get; private set; } = double.NaN;
        /// <summary>
        /// 最大值
        /// </summary>
        public double Max { get; private set; } = double.NaN;
        /// <summary>
        /// 最小值
        /// </summary>
        public double Min { get; private set; } = double.NaN;

        public string Sigma2Text { get; private set; }
        public string MaxText { get; private set; }
        public string MinText { get; private set; }

        /// <summary>
        /// 混合数
        /// </summary>
        public int Mix { get; private set; } = 2;

        /// <summary>
        /// 测量时间
        /// </summary>
        public DateTime Time { get; private set; }

        /// <summary>
        /// 旋转方向 是反向
        /// </summary>
        public bool IsBackw { get; private set; }

        /// <summary>
        /// 旋转1周的时间
        /// </summary>
        public TimeSpan RPeriod { get; private set; }
        /// <summary>
        /// 数据库中的序号
        /// </summary>
        public long Id { get; private set; } = 1231221;
        public int OrgBoltNo { get; private set; } = -100;

        #endregion

        List<Brush> AreaColors;

        ChartValues<double> Values = new ChartValues<double>();
        ChartValues<double> Values2 = new ChartValues<double>();


        BlowingFixProfileParam profileParam;
        ScanGraphItemParam graphparam;
        IBulkDbService bulkDb;
        Lc_ScanData scanData;
        Db_Profile profile;
        IBlowingService blowingService;
        ParamDictionary paramDictionary;
        public ScanGraph2Vm()
        {
            #region 与数据无关界面参数


            AreaColors = new List<Brush>();
            AreaColors.AddRange(FLY.ControlLibrary.Themes.Styles.AreaColors);

            YFormatter = (y) =>
            {
                if (graphparam.IsPercent)
                {
                    if (Target > 0)
                    {
                        if (y == Target)
                        {
                            return y.ToString("F1");
                        }
                        else
                        {
                            double percent = 100.0 * (y - Target) / Target;
                            if (percent > 0)
                            {
                                return $"+{Math.Abs(percent):F1}%";
                            }
                            else
                            {
                                return $"-{Math.Abs(percent):F1}%";
                            }
                        }
                    }
                    else
                    {
                        return y.ToString("F1");
                    }
                }
                else
                {
                    return y.ToString("F1");
                }
            };

            var Mapper = Mappers.Xy<double>()
                .X((value, index) =>
                {
                    return index + 1;
                })
                .Y(value => value)
                .Fill((value) =>
                {
                    if (value > Tolerance2YMax)
                    {
                        return AreaColors[0];
                    }
                    else if (value > ToleranceYMax)
                    {
                        return AreaColors[1];
                    }
                    else if (value >= ToleranceYMin)
                    {
                        return AreaColors[2];
                    }
                    else if (value >= Tolerance2YMin)
                    {
                        return AreaColors[1];
                    }
                    else
                    {
                        return AreaColors[0];
                    }
                });

            var Mapper2 = Mappers.Xy<double>()
                .X((value, index) =>
                {
                    return index + 1;
                })
                .Y(value => value);

            #endregion

            Series = new SeriesCollection
            {
                new Column2Series
                {
                    Title = "扫描",
                    Values = Values,
                    Stroke = new SolidColorBrush(Colors.Black),
                    StrokeThickness = 1,
                    PointGeometry = null,
                    Configuration = Mapper
                },
                new LineSeries
                {
                    Title = "曲线图",
                    Values = Values2,
                    StrokeThickness = 3,
                    Stroke = new SolidColorBrush(Colors.DarkBlue),
                    Fill = new SolidColorBrush(Colors.Transparent),
                    PointGeometry = null,
                    Configuration = Mapper2
                }
            };
            ((Series)Series[0]).SetBinding(Column2Series.YAxisCrossingProperty, new Binding("Target") { Source = this });
        }

        public void Init(
            ScanGraphItemParam graphparam,
            IBulkDbService bulkDb,
            IBlowingService blowingService,
            ParamDictionary paramDictionary,
            BlowingFixProfileParam profileParam
            )
        {
            this.graphparam = graphparam;
            this.bulkDb = bulkDb;
            this.blowingService = blowingService;
            this.paramDictionary = paramDictionary;
            this.profileParam = profileParam;
            Init();
        }
        void Init()
        {
            Misc.BindingOperations.SetBinding(graphparam, nameof(graphparam.Mix), this, nameof(Mix));
            paramDictionary.SetBinding(this, nameof(XStep), 5);

            UpdateY();
            UpdateX();

            Values.Clear();

            Id = -1;
            UpdateId();

            bulkDb.PropertyChanged += BulkDb_PropertyChanged;
            bulkDb.TempFrameChanged += BulkDb_TempFrameChanged;
            graphparam.PropertyChanged += Graphparam_PropertyChanged;
            profileParam.PropertyChanged += ProfileParam_PropertyChanged;
            this.PropertyChanged += ScanGraphVm_PropertyChanged;

            UpdateTemp();
        }

        private void BulkDb_TempFrameChanged(object sender, EventArgs _e)
        {
            BulkDbTempFrameChangedEventArgs e = (BulkDbTempFrameChangedEventArgs)_e;

            if (e.D == null)
            {
                Values2.Clear();
                Average2 = double.NaN;
            }
            else
            {
                int count = e.D.Count();
                int startindex = e.StartIndex;
                int endindex = startindex + count - 1;
                if (Values2.Count <= endindex)
                {
                    //先填充null
                    int dc = endindex + 1 - Values2.Count;
                    double[] array = new double[dc];
                    for (int i = 0; i < dc; i++)
                        array[i] = double.NaN;
                    Values2.AddRange(array);
                }
                for (int i = 0; i < count; i++)
                    Values2[startindex + i] = e.D[i];

                Average2 = Values2.AverageNoNull();

            }
        }

        private void ScanGraphVm_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(Average))
            {
                UpdateTarget();
                UpdateMaxMinText();
            }
            else if ((e.PropertyName == nameof(Max)) || (e.PropertyName == nameof(Min)) || (e.PropertyName == nameof(Sigma2)))
            {
                UpdateMaxMinText();
            }
            else if (e.PropertyName == nameof(Average2))
            {
                UpdateTarget();
            }
            else if (e.PropertyName == nameof(Target))
            {
                if (graphparam.IsPercent)
                {
                    NotifyPropertyChanged(nameof(YFormatter));
                }
                UpdateY2();
            }
            else if (e.PropertyName == nameof(TolerancePercent)) 
            { 
                UpdateY2();
            }
        }

        private void ProfileParam_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(profileParam.Target))
            {
                UpdateOrgTarget();
            }
            else if (e.PropertyName == nameof(profileParam.TolerancePercent)) 
            {
                UpdateOrgTarget();
            }
        }

        private void BulkDb_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(IBulkDbService.LastScanDataId))
            {
                UpdateId();
            }
            else if (e.PropertyName == nameof(FObjBase.FObjServiceClient.IsConnected))
            {
                if ((bulkDb as FObjBase.FObjServiceClient).IsConnected)
                {
                    UpdateId();
                    UpdateTemp();
                }
            }
        }

        private void Graphparam_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(graphparam.YRangePercent))
            {
                UpdateTarget();
                UpdateY2();
            }
            else if (e.PropertyName == nameof(graphparam.Number))
            {
                UpdateId();
            }
            else if (e.PropertyName == nameof(graphparam.IsPercent))
            {
                NotifyPropertyChanged(nameof(YFormatter));
                UpdateTarget();
                UpdateMaxMinText();

            }
            else if (e.PropertyName == nameof(graphparam.IsAutoTarget))
            {
                UpdateTarget();
            }
        }


        void UpdateOrgTarget()
        {
            if (this.profile != null)
            {
                if (this.bulkDb.LastProfileId != this.profile.ID)
                {
                    //使用 profile 数据显示
                    TolerancePercent = this.profile.TolerancePercent;
                    OrgTarget = this.profile.Target;
                }
                else
                {
                    //使用 profileParam 的参数显示
                    TolerancePercent = this.profileParam.TolerancePercent;
                    OrgTarget = this.profileParam.Target;
                }

            }
            else
            {
                //使用 profileParam 的参数显示
                TolerancePercent = this.profileParam.TolerancePercent;
                OrgTarget = this.profileParam.Target;
            }
        }

        void UpdateY()
        {
            UpdateOrgTarget();
            UpdateTarget();
            UpdateY2();
        }

        void UpdateTarget() {

            PollModule.Current.Poll_JustOnce(() =>
            {
                double target = OrgTarget;
                if (graphparam.IsPercent || graphparam.IsAutoTarget)
                {
                    if (!double.IsNaN(Average))
                        target = Average;
                }

                double tolerance = TolerancePercent * target;

                if (tolerance < 0.1)
                    tolerance = 0.1;

                if (!double.IsNaN(Average2))
                {
                    //tempdata 不在显示范围内, 调整 target 为 Average2
                    if (Math.Abs(Average2 - target) > tolerance * graphparam.YRangePercent)
                    {
                        target = Average2;
                    }
                }

                Target = target;
            }, this, MARKNO_TARGET);
        }

        void UpdateY2()
        {
            PollModule.Current.Poll_JustOnce(() =>
            {
                double target = Target;
                double tolerancePercent = TolerancePercent;

                double tolerance = target * tolerancePercent;
                if (tolerance < 0.1)
                    tolerance = 0.1;

                ToleranceYMax = target + tolerance;
                ToleranceYMin = target - tolerance;
                Tolerance2YMax = target + tolerance * 2;
                Tolerance2YMin = target - tolerance * 2;

                YMax = target + tolerance * graphparam.YRangePercent;
                YMin = target - tolerance * graphparam.YRangePercent;
            }, this, MARKNO_Y);
        }
        void UpdateX()
        {
            XMin = 0;
            XMax = this.blowingService.NBolts + 1;
        }

        void UpdateTemp() 
        {
            this.bulkDb.GetTempFrame((asyncContext, retData) =>
            {
                BulkDbTempFrameChangedEventArgs e = retData as BulkDbTempFrameChangedEventArgs;
                BulkDb_TempFrameChanged(null, e);
            }, this);
        }
        void UpdateId()
        {
            Update4IdChanged();
        }

        void UpdateMaxMinText()
        {
            PollModule.Current.Poll_JustOnce(() =>
            {
                if (graphparam.IsPercent && (Average > 0))
                {
                    MaxText = $"+{((Max - Average) / Average) * 100.0:F1}%";
                    MinText = $"-{((Average - Min) / Average) * 100.0:F1}%";
                    Sigma2Text = $"{(Sigma2 / Average) * 100.0:F1}%";
                }
                else
                {
                    MaxText = $"{Max:F1}";
                    MinText = $"{Min:F1}";
                    Sigma2Text = $"{Sigma2:F1}";
                }
            }, this, MARKNO_TEXT);
        }


        void Update4IdChanged2()
        {
            Values.Clear();
            if (this.scanData == null)
            {
                Time = DateTime.MinValue;
                Id = -1;
                OrgBoltNo = -100;
                return;
            }
            else
            {
                Time = this.scanData.Time;
                Id = this.scanData.ID;
                OrgBoltNo = this.scanData.OrgBoltNo;
                IsBackw = this.scanData.IsBackw;
                RPeriod = this.scanData.RPeriod;
            }
            
            double[] values = this.scanData.Thicks;
            Values.AddRange(values);

            var list = from v in Values where !double.IsNaN(v) select v;

            if (list.Count() > 0)
            {
                Average = list.Average();
                Max = list.Max();
                Min = list.Min();
                Sigma2 = list.Sigma() * 2;
            }
            else
            {
                Average = double.NaN;
                Max = double.NaN;
                Min = double.NaN;
                Sigma2 = double.NaN;
            }
        }
        void Update4IdChanged()
        {
            if (graphparam.Number > 0 && bulkDb.LastScanDataId == Id)
            {
                //不需要更新！！！！！！！
            }

            bulkDb.GetFrame(
                new Pack_GetFrameRequest() { Id = graphparam.Number, Mix = graphparam.Mix },
                (object AsyncContext, object retData) =>
                {
                    Pack_GetFrameReponse reponse = retData as Pack_GetFrameReponse;

                    //if (reponse.scanData != null)
                    {
                        this.scanData = reponse.scanData;
                        this.profile = reponse.profile;
                    }
                    
                    Update4IdChanged2();
                    UpdateOrgTarget();
                }, this);
        }

        void NotifyPropertyChanged(string propertyName) {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
        public event PropertyChangedEventHandler PropertyChanged;
    }
}
