﻿using FLY.ControlLibrary;
using FLY.Thick.Blowing.Common;
using FLY.Thick.Blowing.IService;
using FLY.Thick.Blowing.Server.Model;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;

using System.Text;
using System.Threading.Tasks;
using System.Windows.Data;

namespace FLY.Thick.Blowing.UI.Fix.Client.UiModule
{
    public class ScanGraphCircularVm : INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged;


        public int Target { get; set; } = 10000;
        public int Alarm { get; set; } = 300;

        public string Info => $"{Time:MM-dd HH:mm:ss} ID:{Id} 混合数:{Mix}";
        /// <summary>
        /// 测量时间
        /// </summary>
        public DateTime Time { get; private set; }
        /// <summary>
        /// 数据库中的序号
        /// </summary>
        public long Id { get; private set; } = -1;
        public int OrgBoltNo { get; private set; } = 1;
        public int Mix { get; private set; } = 1;

        ScanGraphCircularParam graphparam;
        IBulkDbService bulkDb;
        Lc_ScanData scanData;
        Db_Profile profile;
        IBlowingService blowingService;
        BlowingFixProfileParam profileParam;

        ObservableCollection<int> Values = new ObservableCollection<int>();

        public ScanGraphCircularVm()
        {

        }


        public void Init(
            ScanGraphCircularParam graphparam,
            IBulkDbService bulkDb,
            IBlowingService blowingService,
            BlowingFixProfileParam profileParam,
            GraphScanCircular graph
            )
        {
            this.graphparam = graphparam;
            this.bulkDb = bulkDb;
            this.blowingService = blowingService;
            this.profileParam = profileParam;

            Misc.BindingOperations.SetBinding(profileParam,
                new string[] {
                                nameof(profileParam.Target),
                                nameof(profileParam.TolerancePercent) },

                () => {
                    this.Target = (int)(profileParam.Target * 100);
                    this.Alarm = (int)(profileParam.TolerancePercent * this.Target);
                });

            graph.Title = "环型图";

            graph.SetBinding(GraphScanCircular.BoltNoWithManual1stProperty, new Binding(nameof(ScanGraphCircularParam.BoltNoWithManual1st)) { Source = graphparam });
            graph.SetBinding(GraphScanCircular.ManualCntProperty, new Binding(nameof(ScanGraphCircularParam.ManualCnt)) { Source = graphparam });
            graph.SetBinding(GraphScanCircular.MPIsRightProperty, new Binding(nameof(ScanGraphCircularParam.MPIsRight)) { Source = graphparam });

            //TODO
            graph.SetBinding(GraphScanCircular.BoltCntProperty, new Binding(nameof(blowingService.NBolts)) { Source = blowingService });
            graph.SetBinding(GraphScanCircular.OrgBoltNoProperty, new Binding(nameof(blowingService.OrgBoltNo)) { Source = blowingService });


            graph.SetBinding(GraphScanCircular.Title2Property, new Binding(nameof(Info)) { Source = this });
            graph.SetBinding(GraphScanCircular.TargetProperty, new Binding(nameof(Target)) { Source = this });
            graph.SetBinding(GraphScanCircular.AlarmProperty, new Binding(nameof(Alarm)) { Source = this });

            graph.DataSource = Values;

            graph.ShowSettingDialogEvent += new EventHandler(delegate (object sender, EventArgs e)
            {
                ScanGraphCircularConfig window = new ScanGraphCircularConfig();
                window.Init(this.graphparam);
                window.Owner = FLY.ControlLibrary.COMMON.GetWindow(graph);
                window.ShowDialog();
            });


            Misc.BindingOperations.SetBinding(graphparam, nameof(graphparam.Mix), this, nameof(Mix));


            Id = -1;
            UpdateId();

            bulkDb.PropertyChanged += BulkDb_PropertyChanged;
            graphparam.PropertyChanged += Graphparam_PropertyChanged;
        }

        private void Graphparam_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(graphparam.Mix))
            {
                UpdateId();
            }
            else if (e.PropertyName == nameof(graphparam.Number))
            {
                UpdateId();
            }
        }
        private void BulkDb_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(bulkDb.LastScanDataId))
            {
                UpdateId();
            }
            else if (e.PropertyName == nameof(FObjBase.FObjServiceClient.IsConnected))
            {
                if ((bulkDb as FObjBase.FObjServiceClient).IsConnected)
                {
                    UpdateId();
                }
            }
        }
        void UpdateId()
        {
            Update4IdChanged();
        }
        void Update4IdChanged()
        {
            if (graphparam.Number > 0 
                && bulkDb.LastScanDataId == Id 
                && Mix == graphparam.Mix)
            {
                //不需要更新！！！！！！！
                return;
            }

            bulkDb.GetFrame(
                new Pack_GetFrameRequest() { Id = graphparam.Number, Mix = graphparam.Mix },
                (object AsyncContext, object retData) =>
                {
                    Pack_GetFrameReponse reponse = retData as Pack_GetFrameReponse;

                    if (reponse.scanData != null)
                    {
                        this.scanData = reponse.scanData;
                        this.profile = reponse.profile;
                    }

                    Update4IdChanged2();

                }, this);
        }
        void Update4IdChanged2()
        {
            Values.Clear();
            if (this.scanData == null)
            {
                Time = DateTime.MinValue;
                Id = -1;
                return;
            }
            else
            {
                Time = this.scanData.Time;
                Id = this.scanData.ID;
                OrgBoltNo = this.scanData.OrgBoltNo;
            }

            double[] values = this.scanData.Thicks;
            Values.Clear();
            for (int i=0;i< values.Count();i++)
                Values.Add((int)(values[i]*100));
        }


        void NotifyPropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
    }
}
