﻿using FLY.ControlLibrary;
using FLY.Thick.Blowing.Common;
using FLY.Thick.Blowing.IService;
using Misc;
using OfficeOpenXml.ConditionalFormatting;
using OfficeOpenXml.FormulaParsing.Excel.Functions.Text;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Data;

namespace FLY.Thick.Blowing.UI.Fix.Client.UiModule
{
    public class TrendGraphVm : INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged;
        public int Target { get; set; }
        public int Alarm { get; set; }

        ObservableCollection<GraphRangeData> Values = new ObservableCollection<GraphRangeData>();


        public string Info => $"{Time:MM-dd HH:mm:ss} ID:{Id} 混合数:{Mix}";
        /// <summary>
        /// 测量时间
        /// </summary>
        public DateTime Time { get; private set; }
        /// <summary>
        /// 数据库中的序号
        /// </summary>
        public long Id { get; private set; } = -1;
        public int Mix { get; private set; } = 1;

        FLY.ControlLibrary.GraphRange graph;
        TrendGraphItemParam graphparam;
            IBulkDbService bulkDb;
            BlowingFixProfileParam profileParam;
        public void Init(
            TrendGraphItemParam graphparam,
            IBulkDbService bulkDb,
            BlowingFixProfileParam profileParam,
            FLY.ControlLibrary.GraphRange graphRange) 
        {
            this.graph = graphRange;
            this.graphparam = graphparam;
            this.bulkDb = bulkDb;
            this.profileParam = profileParam;

            Init();
        }

        void Init() {

            Misc.BindingOperations.SetBinding(profileParam,
                new string[] { 
                    nameof(profileParam.Target),
                    nameof(profileParam.TolerancePercent) },

                () =>{
                    this.Target = (int)(profileParam.Target * 100);
                    this.Alarm = (int)(profileParam.TolerancePercent * this.Target);
                });
            
            graph.Title = "纵向图";

            graph.SetBinding(GraphRange.DataCntProperty, new Binding(nameof(graphparam.Len)) { Source = graphparam});
            graph.SetBinding(GraphRange.IsAutoTargetProperty, new Binding(nameof(graphparam.IsAutoTarget)) { Source = graphparam });
            graph.SetBinding(GraphRange.IsPercentProperty, new Binding(nameof(graphparam.IsPercent)) { Source = graphparam });
            graph.SetBinding(GraphRange.YRangePercentProperty, new Binding(nameof(graphparam.YRangePercent)) { Source = graphparam });
            graph.SetBinding(GraphRange.TypeProperty, new Binding(nameof(graphparam.Type)) { Source = graphparam });
            graph.SetBinding(GraphRange.XAxisTypeProperty, new Binding(nameof(graphparam.XType)) { Source = graphparam });

            graph.SetBinding(GraphRange.TargetProperty, new Binding(nameof(Target)) { Source = this });
            graph.SetBinding(GraphRange.AlarmProperty, new Binding(nameof(Alarm)) { Source = this });


            graph.SetBinding(GraphRange.Title2Property, new Binding(nameof(Info)) { Source = this });

            graph.DataSource = Values;


            graph.ShowSettingDialogEvent += new EventHandler(delegate (object sender, EventArgs e)
            {
                TrendGraphConfig w = new TrendGraphConfig();

                TrendGraphItemParam graphparam_src = this.graphparam;

                w.DataContext = graphparam_src;
                w.SaveEvent += (s, ea) =>
                {
                    TrendGraphParams.Current.Save();
                };
                w.Owner = FLY.ControlLibrary.COMMON.GetWindow(graph);
                w.ShowDialog();
            });
            Misc.BindingOperations.SetBinding(graphparam, nameof(graphparam.Mix), this, nameof(Mix));


            Id = -1;
            UpdateId();

            bulkDb.PropertyChanged += BulkDb_PropertyChanged;
            graphparam.PropertyChanged += Graphparam_PropertyChanged;

        }

        private void Graphparam_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(graphparam.Mix))
            {
                Id = -1;
                UpdateId();
            }
            else if (e.PropertyName == nameof(graphparam.Len))
            {
                UpdateId();
            }
        }
        private void BulkDb_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(bulkDb.LastScanDataId))
            {
                UpdateId();
            }
            else if (e.PropertyName == nameof(FObjBase.FObjServiceClient.IsConnected))
            {
                if ((bulkDb as FObjBase.FObjServiceClient).IsConnected)
                {
                    UpdateId();
                }
            }
        }
        void UpdateId()
        {
            Update4IdChanged();
        }



        void Update4IdChanged()
        {
            int cnt;
            if (Id == -1)
            {
                cnt = graphparam.Len;
                Values.Clear();
            }
            else
            {
                cnt = (int)(bulkDb.LastScanDataId - Id);
                if (cnt > graphparam.Len)
                    cnt = graphparam.Len;
            }

            //限制Values长度
            {
                int remove_cnt = Values.Count() - graphparam.Len;
                for (int i = 0; i < remove_cnt; i++)
                    Values.RemoveAt(0);
            }
            if (cnt == 0)
                return;

            bulkDb.GetTrend(
                new Pack_GetTrendRequest() { Id = 0, Mix = graphparam.Mix, Count = cnt },
                (object AsyncContext, object retData) =>
                {
                    Pack_GetTrendReponse reponse = retData as Pack_GetTrendReponse;

                    //向Values 后面添加数据
                    if (reponse.Values != null && reponse.Values.Count > 0)
                    {
                        //从尾向前排的！！！！

                        for (int i = 0; i < reponse.Values.Count(); i++)
                        {
                            int index = reponse.Values.Count() - 1 - i;
                            Values.Add(new GraphRangeData()
                            {
                                Time = reponse.Values[index].Time,
                                Avg = (int)(reponse.Values[index].Value * 100),
                                Sigma = (int)(reponse.Values[index].Sigma * 100)
                            });
                        }
                        //限制Values长度
                        int remove_cnt = Values.Count() - graphparam.Len;
                        for (int i = 0; i < remove_cnt; i++)
                            Values.RemoveAt(0);

                        Id = reponse.ScanDataId;
                        Time = reponse.Values[0].Time;
                    }

                }, this);
        }
    }


}
